## Copyright (C) 2000 - 2006  SWsoft
## All rights reserved.
## Abstract base class for the engine of payment methods

package HSPC::MT::PM::PluginEngine;

use strict;
use SWsoft::protect;

use Storable qw(freeze thaw);
use Error qw(:try);
use Digest::MD5;
use Digest::SHA;

use HSPC::WebDB qw(atrans);
use HSPC::Logger qw(sw_atrace sw_log_warn sw_log);
use HSPC::Localization qw(string);

use HSPC::PluginToolkit::General qw(throw_exception datetime_gmt_now compare_dates);

##use HSPC::MT::Billing::Setup;
##use HSPC::MT::Billing::Constants qw(:DEFAULT :authcall :pend_types :order_types);

##use HSPC::MT::Core;
use HSPC::MT::Core::Constants qw(:DEFAULT :obj_types);

##use HSPC::MT::Core::Phone;

##use HSPC::MT::EV;
##use HSPC::MT::EV::Constants;

use HSPC::MT::PM::Constants qw(
	SW_PM_STATUS_UNCHECKED SW_PM_STATUS_OK SW_PM_STATUS_DISABLED
	SW_PM_STATUS_DELETED SW_PM_STATUS_EXPIRED
);
use HSPC::MT::PM::Tools;

##use HSPC::MT::SS;

##--------------------------------------------------
## Inheritance
##-------------------------------------------------- 
use HSPC::MT::Abstract;

use vars qw{@ISA};
@ISA = qw{HSPC::MT::Abstract};

##--------------------------------------------------
## class info
##-------------------------------------------------- 
use constant CLASS_INFO => {
	props => {
		'id'          => {col => 'id',          table => 'plugin', type => 'int', key=>'AUTO'},
		'type'        => {col => 'type',        table => 'plugin', type => 'char'},
		'name'        => {col => 'name',        table => 'plugin', type => 'char'},
		'account_no'  => {col => 'account_no',  table => 'plugin', type => 'int' },
		'enabled'     => {col => 'enabled',     table => 'plugin', type => 'int' },
		'recurring'   => {col => 'recurring',   table => 'plugin', type => 'int' },
		'status'      => {col => 'status',      table => 'plugin', type => 'int' },
		'template_id' => {col => 'template_id', table => 'plugin', type => 'char', max_length => 100},
		'customer_ip' => {col => 'customer_ip', table => 'plugin', type => 'char', max_length => 100},
		'expire_date' => {col => 'expire_date', table => 'plugin', type => 'char'},
		'data_id'     => {col => 'data_id',     table => 'plugin', type => 'int' },
		'public_data' => {col => 'public_data', table => 'plugin', type => 'char'},
		'checksum'    => {col => 'checksum',    table => 'plugin', type => 'char'},
		'plugin_type' => { sql => 'LCASE(LEFT(plugin.template_id,INSTR(plugin.template_id, "_")-1))' },

		'account_name' => {col => 'name',      table => 'account', type => 'char'},
		'vendor_id'    => {col => 'vendor',    table => 'account', type => 'char'},
	},
	tables => {
		'plugin' => {name => 'plugin_pm'},
		'account' => {
			name       => 'account',
			rel_table  => 'plugin',
			join_where => 'account.account_no = plugin.account_no',
			read_only  => 1
		}
	}
};

## Description:
##	Default constructor
## Parameters:
##	template_id - id of payment plug-in template
## Returns:
##	object reference
## Throws:
##	None
sub new {
	my $class = shift;
	my %h = (
		template_id => undef,
		@_
	);

	my $self = $class->SUPER::new(%h);
	$self->template_id($h{template_id}) if ($h{template_id});
	$self->{_decrypted} = 0;

	return $self;
}

## Description:
##	This method creates a payment method object and loads
##	data for it from the database.
## Parameters:
##	id - id of requested payment method object
## Returns:
##	object reference
## Throws:
##	HSPC::Exception::Plugin::GeneralError
sub find {
	my $class = shift;

	my $trace = sw_atrace();

	my $self = $class->new();
	$self->load(@_);

	## Load template/request DB to understand what type of
	## plugin (online_payment/bank_transfer) is supported
	## payment method type (Visa/DTAUS)
	my %map_subtype = (
		'op' => '::OnlineAccount',
		'bt' => '::BankAccount',
	);

	my $child = 'HSPC::MT::PM::PluginEngine' . $map_subtype{lc($self->plugin_type)};

	eval "use $child;";
	throw_exception(
		type        => 'error',
		log_message => "Cannot load required MT engine: $@"
	) if ($@);

	$self = $child->new();
	$self->load(@_);

	## Unpack public data
	$self->{public_data} = thaw($self->{public_data}) if $self->{public_data};

	## Check expiration date on load
	if ($self->expire_date > 0) {
		my $current_date = datetime_gmt_now();
		if ( &compare_dates($current_date, $self->expire_date) > 0
			&& $self->status != &SW_PM_STATUS_EXPIRED
			&& $self->status != &SW_PM_STATUS_DELETED
		) {
			$self->status(&SW_PM_STATUS_EXPIRED);
			$self->save( only_public => 1, skip_check => 1 );
			eval {
				HSPC::MT::EV->throw_event(
					event_type_id  => &EV_PM_EXPD,
					provider_id    => $self->vendor_id,
					&SW_OBJTYPE_PM => $self->id,
				);
			};
		}
	}

	return $self;
}

## Description:
##	This method saves values of attributes into the database.
## Parameters:
##	only_public - if set to 1, updates only public attributes.
##		Any changes to secret attributes would be lost. This parameter
##		can be provided only for already saved payment methods.
## Returns:
##	undef
## Throws:
##	HSPC::Exception::Plugin::GeneralError
sub save {
	my $self = shift;
	my %h = (
		only_public => undef,
		skip_check  => undef,
		plugin_id   => undef,
		@_
	);

	my $trace = sw_atrace();

	$h{only_public} = 1 unless $self->secure_data;

	if ( $h{only_public} && !$self->id ) {
		throw_exception(
			type        => 'error',
			log_message => 'New payment method cannot be saved with the only_public flag set'
		);
	}

	unless ($h{only_public}) {
		unless (defined $self->{keep_secure_code}) {
			$self->{keep_secure_code} = HSPC::MT::Billing::Setup->is_save_cvc(
				provider_id => $self->vendor_id
			);
		}

		$self->secure_data($self->template->get_secure_data(
			data             => $self->secure_data,
			keep_secure_code => $self->{keep_secure_code},
		)) if ($self->template->can('get_secure_data'));

		unless ($self->checksum) {
			throw_exception(
				type        => 'error',
				log_message => 'The get_public_number method should be defined in the template'
			) unless $self->template->can('get_public_number');

			my $public_number = $self->template->get_public_number(
				secure_data => $self->secure_data,
				public_data => $self->public_data,
			);

			## Claculate checksum
			throw_exception(
				type        => 'error',
				log_message => 'The get_secure_number method should be defined in the template'
			) unless $self->template->can('get_secure_number');

			my $full_number = $self->template->get_secure_number(
				secure_data => $self->secure_data,
				public_data => $self->public_data,
			);

			my $split_pos = int(length($full_number)/2);
			$full_number = reverse($public_number.substr($full_number,0,$split_pos)).substr($full_number,$split_pos);

			my $md5 = new Digest::MD5;
			my $sha = new Digest::SHA(512);
			$md5->add($full_number);
			$sha->add($md5->hexdigest);
			$self->checksum($sha->hexdigest);
		}
	}

	## Validate
	unless ( $h{skip_check} ) {
		$self->validate( plugin_id => $h{plugin_id} );
		if ( $self->template->can('validate') ) {
			$self->template->validate(
				secure_data => $self->secure_data,
				public_data => $self->public_data,
				expire_date => $self->expire_date,
				type        => $self->type,
			);
		}
	}

	## Encrypt
	$self->_encrypt() unless $h{only_public};

	$self->status(&SW_PM_STATUS_DELETED) unless $self->data_id;

	## Pack public data
	$self->{public_data} = freeze($self->{public_data}) if $self->{public_data};

	my $error;
	eval { $error = $self->SUPER::save };

	throw_exception(
		type        => 'error',
		log_message => 'The save of payment method has failed: ' . ($error || $@)
	) if $error || $@;

	## Need to restore public data back
	$self->{public_data} = thaw($self->{public_data}) if $self->{public_data};

	return undef;
}

## Description:
##	This method loads encrypted structure from the database and decrypts it.
## Parameters:
##	sid - session ID that have temporary secure key installed
## Returns:
##	1 - on sucess
## Throws:
##	HSPC::Exception::Plugin::GeneralError
sub decrypt {
	my $self = shift;
	my %h = (
		sid => undef,
		@_
	);

	my $trace = sw_atrace();

	unless ( $self->{_decrypted} ) {
		throw_exception(
			type           => 'error',
			user_message   => {string_id => 'cannot_decrypt_pm_deleted'},
			vendor_message => {string_id => 'cannot_decrypt_pm_deleted'}
		) if $self->status == &SW_PM_STATUS_DELETED;

		throw_exception(
			type           => 'error',
			user_message   => {string_id => 'cannot_decrypt_pm_no_pdata'},
			vendor_message => {string_id => 'cannot_decrypt_pm_deleted'}
		) unless $self->data_id;

		my $vendor_config = HSPC::MT::PP::PayMethodTools->get_vendor_config(
			account_no => $self->account_no
		);

		throw_exception(
			type           => 'error',
			user_message   => {string_id => 'no_pub_key_found'},
			vendor_message => {string_id => 'no_pub_key_found'}
		) unless ($vendor_config->key1 || $vendor_config->key2);

		my $secure_data_string = '';
		## at this step we do not know - skey for which key is specified
		## if spedified. And even more - skey can be empty and we will try
		## to decrypt by stored keys.
		## In all cases we will try to decrypt by both keys.
		foreach ($vendor_config->key1, $vendor_config->key2) {
		    next unless $_;
			try {
				$secure_data_string = HSPC::MT::SS->get_record(
					record_id => $self->data_id,
					key_id    => $_,
					vendor_id => $self->vendor_id,
					sid       => $h{sid},
				);
			} catch HSPC::Exception::Plugin::SecureKey with {
				my $e = shift;
				## Do nothing here because all keys have to be checked
			};
		}

		throw_exception(
			type         => 'key',
			vendor_message => {string_id => 'cant_decrypt_pcci'}
		) unless $secure_data_string;

		$self->secure_data($self->_deserialize(string => $secure_data_string));
		$self->{_decrypted} = 1;
	}

	$trace->add($self->{_decrypted});
	return $self->{_decrypted};
}

## Description:
##	This method marks payment method as deleted.
##	The secure data is removed from the database as well
## Parameters:
##	None
## Returns:
##	None
## Throws:
##	HSPC::Exception::Plugin::GeneralError
sub delete {
	my $self = shift;

	my $trace = sw_atrace();

	$self->enabled(0);
	$self->recurring(0);
	$self->status(&SW_PM_STATUS_DELETED);

	## Remove private data from the database
	if ( $self->data_id ) {
		HSPC::MT::SS->drop_record(record_id => $self->data_id );
		$self->data_id(0);
	}

	## do not save secure data as it has already been deleted
	return $self->save( only_public => 1, skip_check => 1 );
}

## Description:
##	This method verifies consistence of public data.
## Parameters:
##	None
## Returns:
##	undef
## Throws:
##	HSPC::Exception::Plugin::InvalidParameter
sub validate {
	my $self = shift;
	my %h = (
		plugin_id => undef,
		@_
	);

	my $trace = sw_atrace();

	unless ($self->id) {
		## Check existance of the same payment method
    	my $request = HSPC::DataRequest->new(
			cols => [qw(id account_no status checksum)],
			select_as_hashes => 1
		);

		$request->set_col(
			code      => 'checksum',
			for_where => '=',
			value     => $self->checksum
		);
		$request->set_col(
			code      => 'template_id',
			for_where => '=',
			value     => $self->template_id
		);

		HSPC::MT::PM::PluginEngine->load_data(
			request => $request,
		);

		if ( scalar grep {
				(
					$_->{account_no} == $self->account_no &&
					$_->{status} ne &SW_PM_STATUS_DELETED
				) || (
					$_->{account_no} != $self->account_no &&
					$_->{status} eq &SW_PM_STATUS_DISABLED
				)
			} @{$request->{data}}
		) {
			throw_exception(
				type    => 'parameter',
				message => {string_id => 'paymethod_is_exists'}
			);
		}
	}

	my $allowed_types = HSPC::MT::PM::Tools->pm_allowed_types(
		template_id => $self->template_id,
		vendor_id   => $self->vendor_id,
		plugin_id   => $h{plugin_id},
	);

	my $desc = HSPC::MT::PM::Tools->get_pm_template_description(
		types => [$self->type],
	);

	throw_exception(
		type => 'parameter',
		message => {
			string_id   => 'pm_paytype_not_supported',
			string_args => { type => string($desc->[0]->{type_title_id}) }
		}
	) unless ( grep { $_->[0] eq $self->type} @$allowed_types );

	return undef;
}

## >> class
sub apply_to_doc {
	my $self = shift;
	my %h = (
		sec_env                => undef,
		ar_doc                 => undef,
		is_doc_approved        => undef,
		is_doc_approved_reason => undef,
		plugin_id              => undef,
		@_
	);

	my $trace = sw_atrace();
	
	my $ar_doc = $h{ar_doc} || sw_die("Document is not specified");

	my $account_id      = $ar_doc->account_no();
	my $is_doc_approved = $h{is_doc_approved};

	my $checked_ar_doc_id = $self->can_apply_to_docs( doc_ids => [ $ar_doc->id ] );

	unless ( $checked_ar_doc_id && scalar @$checked_ar_doc_id ) {
 		sw_log("This payment method can not be applied to doc #".$ar_doc->id());
 		return undef;
	}

	my $trans = atrans();
	## do noting if cc already applied to document
	my $ar_doc_rb = HSPC::MT::BillingFactory->find_ar_doc_rb( id => $ar_doc->id() );
	if ($ar_doc_rb) {
		if (
			$ar_doc_rb->method_id() == $self->id()
			&& ( !defined $is_doc_approved
				|| $ar_doc_rb->is_doc_approved() == $is_doc_approved )
		  )
		{
			$trans->commit();
			$trace->add("This paymethod assigned already");
			return undef;
		}
	} else {
		$ar_doc_rb = HSPC::MT::Billing::Ar_doc_rb->new();
	};
	
	my $trans_type = &SW_PENDING_TRANS_TYPE_NEW
		unless ($h{trans_type});

	if ( !$trans_type
		&& $ar_doc->doc_type eq &DT_OR
		&& $ar_doc->order_type() == &SW_BM_OT_RENEW
	) {
		$trans_type = &SW_PENDING_TRANS_TYPE_RECURRING;
	}
	$ar_doc_rb->trans_type( $trans_type );
	$ar_doc_rb->id( $ar_doc->id );
	$ar_doc_rb->authcall_status(&SW_AUTHCALL_STATUS_NONE);
	$ar_doc_rb->change_status(
		is_doc_approved => defined($is_doc_approved) ? $is_doc_approved : 1,
		comment         => 'Process Apply to Document.'
			. ($h{is_doc_approved_reason} 
				? ' Reason: ' . $h{is_doc_approved_reason}
				: '')
	);
	$ar_doc_rb->trycount(0);
	$ar_doc_rb->lasterrorcode(0);
	
	$ar_doc_rb->method_id( $self->id );
	$ar_doc_rb->plugin_id($h{plugin_id});
	$ar_doc_rb->save();
	$trans->commit();

	$trace->add("ok");
	return undef;
}

## Abstract method to check can this paymethod be applied
## to proveded documents. This method should be reloaded
## futher during inheratance process
## Parameters:
##	=>> doc_ids - a reference to an array with document IDs
##	<< a reference to an array with allowed document IDs
sub can_apply_to_docs {
	my $self = shift;
	my %h = (
		doc_ids => undef,
		pay_recurrently => undef,
		@_
	);

	## Currently no limitations

	return $h{doc_ids};
}

## This method should be called after successful transaction.
## It performs all necessary actions on payment method for this event
## Parameters:
##  None
## Returns:
##  Undef
sub transaction_approved {
	my $self = shift;
	my %h = (
		only_public => 1,
		@_
	);

	my $trace = sw_atrace();

	$self->status(&SW_PM_STATUS_OK)
		if $self->status == &SW_PM_STATUS_UNCHECKED;

	my $skey = HSPC::MT::PP::PayMethodTools->get_vendor_skey(
		vendor_id => $self->vendor_id()
	);

	$self->save( only_public => ( $h{only_public} || !$skey ) );

	return undef;
}

## This method should be called after failed transaction.
## It performs all necessary actions on payment method for this event
## Parameters:
##  None
## Returns:
##  Undef
sub transaction_declined {
	my $self = shift;
	my $trace = sw_atrace();
	my %h = (
		only_public => 1,
		@_
	);
	my $only_public = $h{only_public};

	if ( HSPC::MT::Billing::Setup->disable_cc_if_declined(
		provider_id => $self->vendor_id()
	)) {
		$self->status(&SW_PM_STATUS_DISABLED);
	}
	$self->save( only_public => $only_public );

	return undef;
}

## Description:
##	Getter that helps in case of just created object (not loaded)
##	and undefined vendor_id.
## Parameters:
##	None
## Returns:
##	Vendor ID
## Throws:
##	None
sub vendor_id {
	my $self = shift;

	my $vendor_id = $self->{vendor_id};

	if ( !$vendor_id && $self->account_no ) {
		$vendor_id = HSPC::MT::Core->find_account_vendor_id($self->account_no);
	}
	return $vendor_id;
}

## Description:
##	Setter/Getter for a payment plug-in template.
## Parameters:
##	None
## Returns:
##	class of a payment plug-in template
## Throws:
##	HSPC::Exception::Plugin::GeneralError
sub template {
	my $self = shift;

	unless ($self->{template}) {
		throw_exception(
			type        => 'error',
			log_message => 'The template_id value should be defined',
		) unless $self->template_id;

		my $template = HSPC::MT::PluginConfig->find_mt_template(
			plugin_type => 'PM',
			template_id => $self->template_id,
		);

		eval "use $template;";

		throw_exception(
			type        => 'error',
			log_message => "Cannot load required MT template: $@"
		) if ($@);

		$self->{template} = $template;
	}

	return $self->{template};
}

## Description:
##	Setter/Getter for a secure data structure.
## Parameters:
##	reference to a secure data structure
## Returns:
##	reference to a secure data structure
## Throws:
##	None
sub secure_data {
	my $self = shift;
	my $data = shift;

	$self->{secure_data} = $data if defined $data;

	return $self->{secure_data};
}

##--------------------------- Private Methods ---------------------------------

## Description:
##	This method encrypts secure structure and saves it into the database.
## Parameters:
##	string - decrypted string that is to be deserialized
## Returns:
##	deserialized object
## Throws:
##	HSPC::Exception::Plugin::GeneralError
sub _deserialize {
	my $self = shift;
	my %h = (
		string => undef,
		@_
	);

	my $string = $h{string} ||
		throw_exception(
			type        => 'error',
			log_message => 'The string parameter must be provided'
		);

	my $data;
	if ($string =~ /^serialize_version=(\d\.\d)/) {
		## get clear freezed string without our "version" tag
		my $substring = substr($string, 21);
		if ( $1 eq '1.0') {
			$data = thaw($substring);
			if ($data->{state} && $data->{state} =~ m/[A-Z]{2}/
				&& ($data->{country} eq 'CA' || $data->{country} eq 'US')
			) {
				$data->{state_uscanada} = $data->{state};
				$data->{state} = '';
			}
			$data->{state_uscanada} = '' unless exists $data->{state_uscanada};

			if (exists $data->{address}) {
				$data->{address1} = $data->{address};
				delete $data->{address};
			}
			if ( $data->{phone} && !exists $data->{phone_src} ) {
				my $phone_src = HSPC::MT::Core::Phone->parse_phone($data->{phone});
				$data->{phone_src} = {
					country_code => $phone_src->{country_code},
					area_code => $phone_src->{area_code},
					number => $phone_src->{number},
					extension => $phone_src->{extension},
				};
			}
			
		} else {
			$data = thaw($substring);
		}
	}
	else {
		## old deserialization method
		my @params = split(/::/, $string);
		$data = {
			card_number  => $params[0],
			holder_name  => $params[1],
			expire_date  => $params[2],
			address      => $params[3],
			cvv          => $params[4],
			zip          => $params[5],
			country      => $params[6],
			city         => $params[7],
			state        => $params[8],
			holder_name2 => $params[9],
			address2     => $params[10],
			phone        => $params[11],
			issue_number => $params[13],
		};
	}

	return $data;
}

## Description:
##	This method encrypts secure structure and saves it into the database.
## Parameters:
##	None
## Returns:
##	undef
## Throws:
##	HSPC::Exception::Plugin::GeneralError
sub _encrypt {
	my $self = shift;

	my $trace = sw_atrace();

	throw_exception(
		type        => 'error',
		log_message => 'Private data is undefined'
	) unless ( $self->secure_data );

	my $vendor_config = HSPC::MT::PM::Tools->get_vendor_config(
		account_no => $self->account_no
	);

	## find keys to encrypt
	throw_exception(
		type           => 'error',
		vendor_message => { string_id => 'no_key_no_save_pm' },
		user_message   => { string_id => 'no_key_no_save_pm' }
	) unless ($vendor_config->key1 || $vendor_config->key2);

	my $string = "serialize_version=1.1" . freeze($self->secure_data);

	my ($key1, $key2);
	if ( $vendor_config->key1 ) {
		$key1 = $vendor_config->key1;
		$key2 = $vendor_config->key2;
	} else {
		$key1 = $vendor_config->key2;
		$key2 = undef;
	}

	my $new_data_id = HSPC::MT::SS->set_record(
		data      => $string,
		key_id    => $key1,
		key_id2   => $key2,
		record_id => $self->data_id || 0,
	);
	$self->data_id($new_data_id);

	$trace->add('Ok');

	return undef;
}

## Description:
##	This method returns a reference to a hash with details of
##  the payment method.
## Parameters:
##	>>  self
## Returns:
##	reference to a hash
sub get_paymethod_info {
    my $self = shift;

	my $result = {};
	$result->{public_data} = $self->public_data;
	$self->decrypt;
	$result->{secure_data} = $self->secure_data;
	$result->{type}        = $self->type;
	$result->{expire_date} = $self->expire_date;


	$result->{name}        = $self->name;
	$result->{account_no}  = $self->account_no;
	$result->{enabled}     = $self->enabled;
	$result->{recurring}   = $self->recurring;
	$result->{status}      = $self->status;
	$result->{customer_ip} = $self->customer_ip;

    return $result;
}

## method for place holder in event manager
sub type_name {
	my $self = shift;
	my $paytype;

	$paytype = HSPC::MT::PM::Tools->get_pm_template_description(
			types => [$self->type]
	) if $self->type;

	return ( scalar @$paytype && $paytype->[0]->{type_title_id} )
		? string( $paytype->[0]->{type_title_id} )
		: $self->type;
}
1;
