#
# Copyright (C) 2000-2004  SWsoft
# All rights reserved.
#
# Presentation engine of payment methods
#

package HSPC::PM::PluginEngine;

use strict;

use SWsoft::protect;
 
use HSPC::Localization qw(:DEFAULT string string_lang current_lang);
use HSPC::WebPage;
use HSPC::DataRequest;

use HSPC::Logger qw(sw_atrace sw_log_warn sw_die sw_log);
use HSPC::WebDB qw(atrans);

use Error  qw(:try);
use HSPC::Exception::Plugin::GeneralError;
use HSPC::Exception::Plugin::SecureKey;
use HSPC::Exception::Plugin::InvalidParameter;

use HSPC::WebSystemLib;

use HSPC::Billing::BM;
use HSPC::Billing::SSUI;
use HSPC::MT::Billing::Constants qw(:DEFAULT :setup :pend_types);
use HSPC::MT::PM::Constants qw(
	SW_PM_STATUS_UNCHECKED
	SW_PM_STATUS_OK
	SW_PM_STATUS_TXT
	SW_PM_STATUS_DELETED
	SW_PM_STATUS_EXPIRED
	SW_PM_STATUS_DISABLED
);
use HSPC::MT::CCP;
use HSPC::MT::Billing::Setup;

use HSPC::MT::Core;
use HSPC::MT::Core::AbstractAccount;
use HSPC::MT::Core::Constants;

use HSPC::MT::GM;
use HSPC::MT::GM::Constants qw(:card_pr_type);

use HSPC::MT::PP;
use HSPC::MT::PM::PluginEngine;
use HSPC::MT::PM::Tools;
use HSPC::MT::PluginConfig;
use HSPC::MT::PluginConfig::PP;
use HSPC::PluginToolkit::General qw(throw_exception get_help_url);
use HSPC::MT::PluginTools qw(set_request_data);

## TO BE REMOVED:
use HSPC::MT::PP::PayMethodTools;
use HSPC::MT::PP::Constants;
use HSPC::EM::Order;
use HSPC::Core::CommentUI;


############################################################################
## Usage: class->new(session_args=>)
## Description: default constructor
## Parameters:
##	session_args =>> session paramaters, see HSPC::PM::SessionArgs
## Return value: new object instance
##
sub new {
	my $class = shift;
	my %h = (
		'session_args' => undef,
		@_,
	);

	sw_die(__PACKAGE__." : session args not defined")
		unless ( ref($h{'session_args'}) 
			&& UNIVERSAL::isa(ref($h{'session_args'}),'HSPC::PM::SessionArgs') );

	my $self = {};
	bless($self,$class);
	$self->{'__session_args'} = $h{'session_args'};

	return $self;
}


############################################################################
## Usage: self->session_args()
## Description: Public getter for session arguments
## Parameters:
##	none
## Return value: reference to HSPC::PM::SessionArgs object instance
## 
sub session_args {
	my $self = shift;

	sw_die(__PACKAGE__." : invalid object method call")
		unless ref($self);

	return $self->{'__session_args'};
}

############################################################################
## Usage: self->teaser_office()
## Description: Main interface method.
## Parameters:   none
## Return value: none
## 
sub teaser_office {
	my $self = shift;
	my $trace = sw_atrace();

	my $sargs = $self->session_args();

	if ( $sargs->action eq 'view') {
		return $self->teaser_view();
	}
	if ( $sargs->action eq 'edit') {
		if (sw_argparam('view_pr_info')) {
			return $self->teaser_skey();
		} elsif (sw_argparam('edit_docs')) {
			return $self->teaser_edit_docs();
		} else {
			return $self->teaser_edit();
		}
	}

	if ( $sargs->action eq 'skey') {
		return $self->process_save_skey();
	}

	if ( $sargs->action eq 'edit_docs') {
		return $self->teaser_edit_docs();
	}
	if ( $sargs->action eq 'pay_doc') {
		return $self->process_update_docs();
	}
	if ( $sargs->action eq 'update') {
		return $self->process_update_paymethod();
	}
	if ( $sargs->action eq 'select_pt') {
		return $self->teaser_template_select();
	}
	if ( $sargs->action eq 'add') {
		return $self->teaser_add();
	}
	if ( $sargs->action eq 'popup') {
		return $self->teaser_accounts_popup();
	}
	if ( $sargs->action eq 'pm_help') {
		return $self->paymethod_help();
	}
	if ( $sargs->action eq 'create') {
		if ( sw_argparam('just_reload') ) {
			$sargs->paymethod_acc_no(&sw_argparam('account_no'));
			return $self->teaser_add();
		} else {
			return $self->process_create_paymethod();
		}
	}
	return $self->teaser_list();
}


## Draws list of all paymethods defined for an account
## This listing inclides Credit Cards paymethod along with Bank Accounts
sub teaser_list {
	my $self = shift;
	my $trace = sw_atrace();

	my $sargs = $self->session_args;
	my $main_page = $sargs->page;

	if ( $sargs->draw_header() ) {
		## Backward compatibility for CP that's not migrated to COMPREP yet
		if ($ENV{COMPREP}) {
			$main_page->title(pages => []);
		} else {
			$self->build_path( step => $self->build_path() );

			$main_page->header();
			$main_page->title(
				'title_id' => 'paymethods_uc',
				'devscreen_id' => $self->get_screen_id( page => 'list' ),
				'refresh_url' => $main_page->{'common_url'},
			);
			$main_page->tab();
		}
	}
	$main_page->statuses();	

	## Redefine these properties of page just to make sure that list method will
	## show correct wording
	$main_page->{'res_itemstitle_id'} = $self->__title_plural_id;
	$main_page->{'res_itemtitle_id'} = $self->__title_single_id,

	my $pm_types = HSPC::MT::PM::Tools->get_pm_template_description(
		plugin_type => $sargs->plugin_type
	);

	my $data_options = ();
	my $filter_options = ();
	push(@$filter_options, [undef, 'all']);
	foreach my $pm_type (@$pm_types) {
		push (@$data_options, [
			$pm_type->{type},
			$pm_type->{image_id} ? "<img_".$pm_type->{image_id}.">" : string($pm_type->{type_title_id})
		]);
		push (@$filter_options, [$pm_type->{type}, $pm_type->{type_title_id}]);
	}

	## Payment method statuses
	my $cc_status = &SW_PM_STATUS_TXT;
	$cc_status = [ map {[$_,string($cc_status->{$_})]} keys %$cc_status ];
	unshift @$cc_status, [undef, string('all_but_deleted')];

	
	## prepare request for data
	my $link_base = $main_page->{common_url}."/act,view/pmethod_id,";
	my $cols = [
		{
			title_id  => 'id',
			code      => 'id',
			type      => 'id',
			sort_def  => 2,
			link_base => $link_base,
			for_search   => 1,
			search_width => 5,
		},
		{
			title_id     => 'name',
			code         => 'name',
			type         => 'char',
			for_search   => 1,
			link_base    => $link_base,
		},
		{
			title_id        => 'type',
			code            => 'type',
			type            => 'options',
			data_options    => $data_options,
			filter_options  => $filter_options,
			for_filter      => 1,
			filter_localize => 1,
			align           => 'center',
		},
		{
			title_id => 'exp_date',
			code     => 'expire_date',
			type     => 'char',
		},
	];
	if ($sargs->is_admin) {
		push @$cols, {
			title_id     => 'accnt_uc',
			code         => 'account_name',
			type         => 'char',
			for_search   => 1,
		};
	}

	if ( $sargs->account_type == &SW_RESELLER
		&& ! $sargs->paymethod_acc_no
	) {
		push @$cols, {
			title_id     => 'accnt_type_uc',
			code         => 'vendor_id',
			for_filter   => 1,
			type         => 'char',
			hide         => 1,
			filter_options => [
				[ $sargs->account_no, string('customer') ],
				[ &SW_HSP_ID, string('reseller') ]
			],
		};
	}

	push @$cols, {
		title_id       => 'status',
		code           => 'status',
		type           => 'char',
		for_filter     => 1,
		filter_options => $cc_status,
		no_wrap        => 1,
		not_escape     => 1,
	}, {
		code => 'account_no',
		type => 'int',
		hide => 1,
	}, {
		title_id => 'allow_payments',
		code     => 'enabled',
		type     => 'bool',
	}, {
		title_id => 'use_4_rec_billing',
		code     => 'recurring',
		type     => 'bool',
	};


	my $request = $main_page->prepare_browse(
		cols => $cols,
	);

	if ( $sargs->account_type == &SW_RESELLER ) {
		if ( $request->value('vendor_id') && $request->value('vendor_id') != $sargs->account_no ) {
			## Resellers` own carts
			$request->set_col( code => 'account_no', for_where => ' = '.$sargs->account_no );
		} else {
			## Customers` carts
			$request->set_col( code => 'vendor_id', for_where => ' = '.$sargs->account_no )
				unless ($request->value('vendor_id'));
			## If called from AM show carts for defined customer
			$request->set_col( code => 'account_no', for_where => ' = '.$sargs->paymethod_acc_no )
				if $sargs->paymethod_acc_no;
		}
	} elsif ($sargs->account_type == &SW_HSP) {
		$request->set_col( code => 'vendor_id', for_where => ' = '.$sargs->account_no );
		$request->set_col( code => 'account_no', for_where => ' = '.$sargs->paymethod_acc_no )
			if $sargs->paymethod_acc_no;
    } elsif ($sargs->account_type == &SW_CUSTOMER) {
        $request->set_col( code => 'account_no', for_where => ' = ' . ($sargs->paymethod_acc_no || $sargs->account_no) );
	}

	if ( $request->value('status') eq '' ) {
		$request->set_col( code => 'status', for_where => '!=' . &SW_PM_STATUS_DELETED );
	}

	if ($sargs->plugin_type) {
		$request->set_col( code => 'plugin_type', for_where => ' = LCASE("'.$sargs->plugin_type.'")');
	}

	HSPC::MT::PM::PluginEngine->load_data ( 
		request => $request,
	);

	my $status_i = $request->get_index_by_code('status');
	my $exdate_i = $request->get_index_by_code('expire_date');

	foreach (@{$request->{data}}) {
		if($status_i >= 0) {
			$_->[$status_i] = $self->colored_status($_->[$status_i]);
		}		
		if($exdate_i >= 0) {
			if ( 0 + $_->[$exdate_i] > 0 ) {
				my ($year,$month) = ($_->[$exdate_i] =~ /^(\d{4})-(\d\d)/);
				$_->[$exdate_i] = "$month/$year";
			} else {
				$_->[$exdate_i] = string('never_uc');
			}
		}		
    }

	my $gm_config = HSPC::MT::PM::Tools->get_vendor_config(
		vendor_id => $sargs->vendor_id()
	);

	unless ( $gm_config->key1 || $gm_config->key2 ) {
		$main_page->add_status( string_id => 'pk_not_stored_cc_disabled' );
		$main_page->statuses();
	} else {
		my $plugins = HSPC::MT::PP->get_pp_plugin_info(
			is_active  => 1,
			is_direct  => 1,
			sub_type   => $sargs->plugin_type,
			vendor_id  => $sargs->vendor_id()
		);

		$main_page->actions_bar( btns => [ {
			title    => string('new_paymethod_title', name => string($self->__title_single_id)),
			icon     => 'new_visa',
			url_ext  => '/act,select_pt'
		}]) if (scalar @$plugins);
	}

	$main_page->search_bar();
	$main_page->errors($main_page);

	$main_page->browse( 
		browse_info		=> $request, 
	);

	$main_page->footer();
	$trace->add("ok");

	return undef;
}

###############################################################################
## END (Visual PayMethod Plugins API)
###############################################################################

##---------------------------------------------------------------------------------

## form to enter secure key and passphrase
## 
##
sub teaser_skey {
	my $self = shift;

	my $trace = sw_atrace();
	my $sargs = $self->session_args;

	unless ( $sargs->paymethod_obj && $sargs->paymethod_obj->id ) {
		sw_die(__PACKAGE__ . ": Paymethod object is undefined");
	}

	my $main_page = $sargs->page;

	my $url_ext = "/act,skey/pmethod_id,".$sargs->paymethod_obj->id;

	$main_page->edit_open(
		url_ext  => $url_ext,
		title_id => 'enter_skey_n_passphrase',
		is_multipart => 1
	); 

	HSPC::Billing::SSUI->secure_key_form(page => $main_page);
	HSPC::Billing::SSUI->secure_phrase_form(page => $main_page);

	my $setup = HSPC::MT::Billing::Setup->new(
		provider_id => $sargs->vendor_id
	);

	$main_page->row_open();	
	$main_page->cell_skip();	
	$main_page->cell_input(
		view_name=>'store_interval',
		value=> &sw_argparam('store_interval') || $setup->store_interval || 1,
		max_length => 2,
		size => 2,
		check_title => string('store_interval_explained'),
	);
	$main_page->cell_text(
		value => string('store_interval_explained'),
		is_aster => 1
	);
	$main_page->row_close();					

	my $buttons = [ {
		title => string('view_private_card_data_uc'),
		name  => 'view_pr_info_with_key',
		type  => 'submit',
		class => 'SWIButton'
	} ]; 

	$main_page->edit_close( upd_id => 0, btns => $buttons );
	return undef;
}


## static method
## check, if secure key passed from input or available in database
## >> class
## >> key_ref - hashref to put key info into
## >= cc - reference to payment method object
## >= vendor_id
## >= account_no
## < undef if failed or key_id if success
sub process_save_skey {
	my $self = shift;
	my $trace = sw_atrace();

	my $sargs = $self->session_args;

	unless ( $sargs->paymethod_obj && $sargs->paymethod_obj->id ) {
		sw_die(__PACKAGE__ . ": Paymethod object is undefined");
	}

	my $main_page = $sargs->page;

	unless ( sw_argparam('item_cancel') ) {
		my $error = HSPC::Billing::SSUI->save_skey_phrase_temp(
			vendor_id => $sargs->vendor_id
		);

		if ( $error ) {
			$main_page->add_error( string => $error	);
			return $self->teaser_skey();
		}
	}

	return $main_page->redirect(
		undef, undef, "/act,view/pmethod_id," . $sargs->paymethod_obj->id
	);
}

use Socket;
use constant WHOIS_LINKBASE => 'http://ws.arin.net/cgi-bin/whois.pl?queryinput=';

sub display_ip {
	my $self = shift;
	my $trace = sw_atrace();

	my $ip = shift;
	return unless $ip;
	my $display_ip =
	  qq[<a href="] . &WHOIS_LINKBASE . qq[$ip" target="_blank">$ip</a>];
	my $name = gethostbyaddr(pack('CCCC', split(/\D/, $ip)), AF_INET);
	$display_ip .= $name ? ' (' . $name . ')' : ' (DNS lookup failed)';
	return $display_ip;
}

sub teaser_view {
    my $self = shift;

	my $trace = sw_atrace();
	my $sargs = $self->session_args;

	my $paymethod = $sargs->paymethod_obj();

	unless ( $paymethod && $paymethod->id ) {
		sw_die(__PACKAGE__ . ": Paymethod object is undefined");
	}

	if ( $paymethod->vendor_id != $sargs->account_no &&
		$paymethod->account_no != $sargs->account_no
	) {
		$ENV{system_obj}->exit_noobject();
	}

	my $main_page = $sargs->page;

	if ( $sargs->draw_header() ) {
		## Backward compatibility for CP that's not migrated to COMPREP yet
		if ($ENV{COMPREP}) {
			$ENV{COMPREP}->node(edit => 1)->{title_hash}->{title} =
				string('paymethod') . ' #' . $paymethod->id;
			$main_page->title(pages => [],
				comment_obj_id=>$paymethod->id,
				comment_obj_type=>$paymethod->obj_type_id()
			);
		} else {
			$main_page->header();
			$main_page->title(
				'title' => string('paymethod')." #".$paymethod->id,
				'devscreen_id' => $self->get_screen_id( page => 'view' ),
			);
			$self->build_path( step => 3 );

			$main_page->tab();
		}

	}
	$main_page->statuses();

	$main_page->view_info_header();

	## draw payment method general settings view screen
	$self->view_form(
		obj              => $paymethod,
		page             => $main_page,
		is_admin         => $sargs->is_admin,
		account_type     => $sargs->account_type,
		effective_acc_no => $sargs->account_no,
	);

## BUTTONS ---------------------------------------------
	my $url_ext = "/act,edit/pmethod_id,".$paymethod->id;
	my @buttons;
	
	## If we have no secure key - show button for enter it
	my $vendor_config = HSPC::MT::PP::PayMethodTools->get_vendor_config(
		vendor_id => $paymethod->vendor_id
	);

	my $stored_key_id = HSPC::MT::PP::PayMethodTools->get_vendor_skey(
		vendor_id => $paymethod->vendor_id
	);

	if ( $sargs->is_admin && !$stored_key_id
		&& ( $vendor_config->key1 || $vendor_config->key2 )
	) {
	  push @buttons, ['view_pr_info', string('view_private_card_data_uc'),  'SWButton','submit'];
	}

	## merge part - one row View documents
	my $setup = HSPC::MT::Billing::Setup->new (provider_id => $paymethod->vendor_id);
	my $is_owner = ($paymethod->account_no == $sargs->account_no);
	my $allow_owner_cc_operations = 0;
	if ($is_owner && $setup 
		&& ($setup->rec_billing_mode == &SW_BM_RECBILL_HUMAN 
			|| $setup->rec_billing_mode == &SW_BM_RECBILL_AUTO))
	{
		$allow_owner_cc_operations = 1;
	}

	## if we set "both manual" - we can assign CC to doc in any rate
	## if we have active online payment plugin we can also let a chance 
	## to assign a card to a document
	my $plugins = HSPC::MT::PP->get_pp_plugin_info(
		is_active  => 1,
		is_direct  => 1,
		sub_type   => $paymethod->plugin_type,
		vendor_id  => $paymethod->vendor_id
	);

	if ( ( $vendor_config->key1 || $vendor_config->key2 ) &&
		( $vendor_config->card_proc_type == SW_GM_CCARD_PRTYPE_MANUAL || @$plugins )
		&& (($is_owner && $allow_owner_cc_operations) || !$is_owner)) 
	{
		push @buttons, ['edit_docs', string('pay_docs_uc'), 'SWButton', 'submit']
			if ( $self->can_pay_docs(paymethod => $paymethod) );
	}
	
	$main_page->view_info_footer();
	## list comments
	my $res = HSPC::Core::CommentUI->list_internal(
		obj_type => $paymethod->obj_type_id(),
		obj_id   => $paymethod->id(),
		is_complete_html => 1,
	);

	if ($res->{STATUS} eq 'OK' && $res->{SIZE}>0 ) {
		$main_page->{edit_out} .= $res->{CONTENT};
	}
	$main_page->view_info_button(show_cancel => 1, url_ext => $url_ext,
		show_delete => ($paymethod->status != &SW_PM_STATUS_DELETED),
		show_edit => ( (!$sargs->is_admin || $stored_key_id) &&
			($paymethod->status != &SW_PM_STATUS_DELETED)
			&& (($is_owner && $allow_owner_cc_operations) || !$is_owner) ),
		right_buttons=>\@buttons);

	return undef;
}

sub teaser_accounts_popup {
	my $self = shift;
	my $trace = sw_atrace();

	my $sargs = $self->session_args;
	my $main_page = $sargs->page;

	my %arg=();

	## show only appropiate account owned by current vendor
	$arg{vendor} = $sargs->vendor_id;

	$main_page->{'common_url'} .= "/act,popup";

	HSPC::Core::CoreUI->teaser_account_popup(
					page      => $main_page,
					form_name => 'item',
					view_name => 'account_no', 
					force_account_type => 1,
					account_type => (
						$ENV{security_obj}->account_type() == SW_HSP
							? [ SW_RESELLER, SW_CUSTOMER ]
							: [SW_CUSTOMER]
					),
					js_code   => "parent.opener.document.getElementById('item').just_reload.value=1;".
						         "parent.opener.document.getElementById('item').submit();",
					%arg,
				);

	return undef;
}

sub paymethod_help {
	my $self = shift;
	my $sargs = $self->session_args;
	my $page = $sargs->page;

	## try to find presentation template
	my $template_id = $sargs->template_id || (
		$sargs->paymethod_obj &&
		$sargs->paymethod_obj->template_id
	);

	my $template = HSPC::MT::PluginConfig->find_presentation_template(
		plugin_type => 'PM',
		template_id => $template_id,
	);

	eval "use $template;";
	throw HSPC::Exception::Plugin::GeneralError(
		log_message => "Can not load presentation engine: $@"
	) if ( $@ );

	$page->{edit_out} = $template->get_help_page();

	return undef;
}

##-----------------------------------------------------------------------------

## >> class 
## => cc - optional CC obj to be used on add form
## => prefill
### + same arguments as for "teaser"
sub teaser_add {
	my $self = shift;

	my $trace = sw_atrace();
	my $sargs = $self->session_args;
	my $main_page = $sargs->page;
	my $paymethod = $sargs->paymethod_obj();

	if (sw_argparam('item_cancel')) {
		return $main_page->redirect();
	}

	unless ( $sargs->template_id ) {
		$main_page->add_error( string_id => 'pls_sel_pmnt_method_type' );
		return $self->teaser_template_select();
	}

	if ( $sargs->draw_header() ) {
		## Backward compatibility for CP that's not migrated to COMPREP yet
		if ($ENV{COMPREP}) {
			$main_page->title(pages => []);
		} else {
			$self->build_path( step => 3 );

			## Find paymethod title
			my $paytype = HSPC::MT::PM::Tools->get_pm_template_description(
				template_id => $sargs->template_id
			);
			
			$main_page->header();
			$main_page->title(
				'title' => string('new_paymethod_title',
					name => string($paytype->[0]->{pm_title})
				),
				'devscreen_id' => $self->get_screen_id(
					page => 'add_'.$paytype->[0]->{plugin_type}
				),
			);
			$main_page->tab();
		}
	}
	$main_page->statuses();

	my $url_ext = "/act,create";

	$main_page->edit_open(url_ext=>$url_ext,disable_header=>1);

	my $show_account_selector = 0;
	if ($sargs->is_admin || ($sargs->account_type eq &SW_RESELLER)) {
		$show_account_selector = 1;
	}

	## draw account general settings edit screen
	my $js_code = $self->add_form(
		page                  => $main_page,
		obj                   => $sargs->paymethod_obj(),
		account_no            => $sargs->paymethod_acc_no(),
		show_account_selector => $show_account_selector,
		prefill               => 1,
		is_admin              => $sargs->is_admin,
		account_type          => $sargs->account_type,
		effective_acc_no      => $sargs->account_no,
		template_id           => $sargs->template_id,
		vendor_id             => $sargs->vendor_id,
	);

	$main_page->edit_close(
		upd_id => 0, cancel_id => 'back',
		btns => [{
			title => string('ok'), name => 'btn_submit', type => 'submit',
			class => 'SWButtonX', check_form => 1
		}],
		code_custom => $js_code 
	);

	$main_page->footer();

	return undef;
}

sub teaser_template_select {
	my $self = shift;

	my $trace = sw_atrace();
	my $sargs = $self->session_args;
	my $main_page = $sargs->page;

	## find all available templates
	my $paymethod_templates = HSPC::MT::PM::Tools->find_paymethod_plugins(
		active      => 1,
		plugin_type => $sargs->plugin_type,
		vendor_id   => $sargs->vendor_id
	);

	if ( scalar @$paymethod_templates == 0 ) {
		throw HSPC::Exception::Plugin::GeneralError(
			log_message => "There is no available templates to add new payment method"
		);
	}

	if (scalar @$paymethod_templates == 1) {
		$main_page->redirect(undef,undef,"/act,add/template_id," . $paymethod_templates->[0]);
	}

	if ( sw_argparam('item_cancel') ) {
		return $main_page->redirect();
	}

	if ( $sargs->draw_header() ) {
		$self->build_path();

		$main_page->header();
		$main_page->title(
			'title_id' => 'select_paymethod_type',
			'devscreen_id' => '02.21',
			'refresh_url' => $main_page->{'common_url'}.$main_page->get_search_url(),
		);
		$main_page->tab();
	}
	$main_page->statuses();

	$main_page->edit_open(
		title_id=>'select_paymethod_type',
		url_ext=>"/act,add"
	);

	foreach my $template_id (@$paymethod_templates) {
		my ($payment_type) = HSPC::MT::PM::Tools->get_pm_template_description(
			template_id => $template_id
		);

		$main_page->edit_view_radio(
			view_name => "template_id",
			title_id  => $payment_type->[0]->{pm_title},
			value     => $template_id,
		);
	}

	
	$main_page->edit_close(upd_id=>'next');
	$main_page->footer();
	return undef;
}

## > account status
sub colored_status {
  my $trace = sw_atrace();

	my $class = shift;
	my $value = shift;
	my $custom = $ENV{HSPC_SYSTEM}->ui_palette;

	## by default
	my ($color,$status_icon) = ();

	if ($value == &SW_PM_STATUS_OK) {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id=>'alert_ok');
		$color = $custom->{Status};
	} elsif($value == &SW_PM_STATUS_UNCHECKED || $value == &SW_PM_STATUS_DELETED ) {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id=>'alert_inprogress');
	} else {
		$status_icon = $ENV{HSPC_SYSTEM}->ui_image_src(id=>'alert_error');
		$color = $custom->{Error};
	}

	my $st = string(&SW_PM_STATUS_TXT->{$value});
	$st = qq[<font color="$color">$st</font>] if defined $color;
	return $status_icon.'&nbsp;'.$st;
}

## check if we can pay the documents
## > class
## => paymethod
## => is_admin
sub can_pay_docs {
	my $self = shift;
	my $trace = sw_atrace();

	my %h = (
		paymethod => undef,
		@_
	);

	my $sargs = $self->session_args;

	sw_die("The paymethod parameter must be provided") unless ($h{paymethod} && $h{paymethod}->id);

	if( $h{paymethod}->status == &SW_PM_STATUS_DISABLED ||
		$h{paymethod}->status == &SW_PM_STATUS_DELETED
	) {
		$trace->add("[R1] Paymethod is deleted or disabled");
		return 0;
	}

	unless( $h{paymethod}->status == &SW_PM_STATUS_OK
		|| $h{paymethod}->status == &SW_PM_STATUS_UNCHECKED
		|| $sargs->is_admin
	) {
		$trace->add("[R2] Status of paymethod is not OK or UNCHECKED");
		return 0;
	}

	## Find plugin IDs that support current payment method type
	my $plugin_ids = HSPC::MT::PM::Tools->get_plugins_supported_by_paytype(
		type      => $h{paymethod}->type,
		vendor_id => $h{paymethod}->vendor_id,
	);

	unless ( scalar @{$plugin_ids} ) {
		$trace->add("[R3] No active plugins that support this payment type (" . $h{paymethod}->type . ") found");
		return 0;
	}

	my $request = new HSPC::DataRequest
		(cols=>[{code=>'ar_doc_id',type=>'id'}]);

	$request->set_where('doc_type', [&DT_IN, &DT_OR, &DT_DA]);
	$request->set_col(code=>'doc_balance',for_where=>'>0');
	$request->set_where('doc_status',[&DS_OPEN,&DS_OVERDUE]);
	$request->set_where('batch_id', 0 );
	$request->set_where('is_doc_approved', 1 );

	HSPC::MT::Billing::Ar_doc_list->ar_doc_cacp_list(
		$request,$h{paymethod}->account_no()
	);

	unless ( scalar @{$request->data()} ) {
		$trace->add("[R4] No unpaid documents found");
		return 0;
	}

	my @doc_ids = map { $_->[0] } @{$request->data()};
	my $allowed_docs = $h{paymethod}->can_apply_to_docs( doc_ids => \@doc_ids );

	if ( scalar @$allowed_docs ) {
		$trace->add("[R5] There are unpaid documents that can be paid");
		return 1;
	}
	$trace->add("[R6] No unpaid documents that can be paid found");
	return 0;
}

##-----------------------------------------------------------------------------
##  MERGE PART
## >>  class
## =>> page
## =>> obj : reference to payment method object
## =>> is_dmin
sub edit_docs {
	my $self = shift;
	my %h = (
		page => undef,
		obj => undef,
		@_
	);
	my $trace = sw_atrace();

	my $old_page = $h{page} || &sw_die('page=> undefined');
	my $cc = $h{obj} || &sw_die('obj=> undefined');

	my $page = new HSPC::WebPage();
	$page->{res_itemstitle_id} = 'docs';
	$page->{common_url} = $old_page->{common_url};
	## find unpaid orders/invoice
	my $request = $page->prepare_browse(
		cols => [
			{
				code      => 'id',
				type      => 'id',
				no_sort   => 1,
				for_sort  => 1,
				for_check => 1,
				title_id  => 'id'
			},
			{
				code     => 'description',
				type     => 'q',
				title_id => 'desc',
				no_sort  => 1
			},
			{
				code     => 'doc_date',
				type     => 'datetime',
				title_id => 'doc_date_uc',
				no_sort  => 1
			},
			{
				code       => 'doc_total',
				type       => 'money',
				title_id   => 'total',
				no_sort    => 1,
				vendor_col => 'vendor',
			},
			{
				code       => 'doc_balance',
				type       => 'money',
				title_id   => 'balance',
				no_sort    => 1,
				vendor_col => 'vendor',
			},
			{
				code     => 'method_id',
				type     => 'char',
				title_id => 'paymethod_uc',
				no_sort  => 1,
				not_escape => 1,
			},
			{
				code     => 'vendor',
				type     => 'int',
				hide     => '1',
			}
		]
	);
	$request->set_where('account_no',$cc->account_no());
	$request->set_col(code=>'doc_status', for_where=>1,value=>[&DS_OPEN,&DS_OVERDUE]);
	$request->set_col(code=>'doc_type',   for_where=>1,value=>[&DT_IN,&DT_OR,&DT_DA,&DT_PD]);
	$request->set_col(code=>'doc_balance',for_where=>'>',value=>0);
	$request->set_col(code=>'is_doc_approved',for_where=>'1',value=>1);
	
	## ar_doc_rb
	my $join = 'left join ar_doc_rb as adr on adr.ar_doc_id = ar_doc.ar_doc_id ';
	my $gate_cols = {method_id=>'adr.method_id', is_doc_approved => 'adr.is_doc_approved'};

	sw_log ("Loading data...");
	HSPC::MT::Billing::Ar_doc->load_data(
		request  =>$request,
		'join'   =>$join,
		gate_cols=>$gate_cols
	);
	my @data = @{$request->{data}};

	$ENV{system_obj}->exit_noobject() unless ( scalar(@data) );

	my @ready_data = ();
	my @checklist_selected;
	my $id_index = $request->get_index_by_code('id');
	my $pay_id_index = $request->get_index_by_code('method_id');

	foreach my $rec (@data) {
		my $can_pay_doc = $cc->can_apply_to_docs( doc_ids => [$rec->[$id_index]] );
		next unless ( $can_pay_doc && scalar @$can_pay_doc );

		my $res = HSPC::MT::BM->is_doc_permitted_for_payment( doc_id => $rec->[$id_index] );
		## if document is permitted for payment - copy it to ready array
		if ( $res->{code} == SW_RESCODE_OK ) {
			my $method_id = $rec->[$pay_id_index];
			push @checklist_selected, $rec->[$id_index] if ($cc->id() == $method_id);

			if ( $method_id ) {
				## Credit card exists
				my $card = HSPC::MT::PP->find_paymethod( id => $method_id );
				if ( $card ) {
					my ($type) = HSPC::MT::PM::Tools->get_pm_template_description(
						types => [$card->type]
					);
					$rec->[$pay_id_index] = $card->name().'&nbsp;';
					$rec->[$pay_id_index] .= $ENV{HSPC_SYSTEM}->ui_image_src(
						id    => $type->[0]->{image_id},
						align => 'MIDDLE'
					)."<br>";
					$rec->[$pay_id_index] .= $self->colored_status($card->status());
				} else {
					sw_log_warn("Can`t load payment method #$method_id defined as paymethod for document #" . $rec->[$id_index]);
				}
			}
			push @ready_data, $rec;
		}
	}

	## now in "request" variable only "DA" data.
	## we have to fill it with correct data contains both (OR, IN) and DA
	$request->{data} = \@ready_data;
	
	## Action buttons
	my @checklist_buttons = (
		{
			name     => 'btn_submit',
			title_id => 'apply_pm_to_docs',
			icon     => 'icon_bm',
		},
	);	
	
	$page->browse(
		browse_info           =>$request,
		checkform_name        => 'item',
		hide_paging           =>1,
		checklist_form_noopen =>0,
		checklist_form_noclose=>0,
		checklist_selected   => [@checklist_selected],
		checklist_buttons 	 => \@checklist_buttons,		
		checklist_full_url => $page->{common_url}.$h{act_url_ext},
	);

	$old_page->{edit_out} .= $page->{edit_out};

}## edit_docs

sub process_update_docs {
	my $self = shift;
	my $trace = sw_atrace();

	my $sargs = $self->session_args;

	unless ( $sargs->paymethod_obj ) {
		sw_die(__PACKAGE__ . ": Paymethod object is undefined");
	}

	my $paymethod = $sargs->paymethod_obj;
	## find unpaid orders/invoice/debit_adjustments
	my $request = new HSPC::DataRequest(
		cols=>[
		{code=>'id',type=>'id'},
		]
	);
	HSPC::MT::Billing::Ar_doc_list->ar_doc_unpaidlist(
		$request,$paymethod->account_no(),
		request  =>$request
	);

	my $sec_env = &get_sec_env();
	my $error;
	my @ids = @{ sw_argparam( 'id', 'list' ) };
	my $user_call = $ENV{security_obj}->account_type == &SW_CUSTOMER ? 1 : 0;
	foreach my $rec (@{$request->{data}}) {
		if ( grep {$rec->[0] eq $_} @ids ) {
			my $doc = HSPC::MT::BillingFactory->ar_doc_find(id => $rec->[0]);
			try {
				my $res = HSPC::MT::PP->payment(
					doc              => $doc,
					sec_env          => $sec_env,
					paymethod        => $paymethod,
					billing_details  => {
							skip_if_3ds_require => 1,
						}
				);
			} catch HSPC::Exception::Plugin::SecureKey with {
				if ($user_call) {
					my $res = HSPC::MT::CCP->save_pending_document(
						order                  => $doc,
						method_id              => $paymethod->id(),
						use_stored_card        => defined $paymethod,
						is_doc_approved        => 0,
						is_doc_approved_reason => 'SecureKey error',
						trans_type             => &SW_PENDING_TYPE_UPGRADE
					);
				} else  {
					$error = shift;
					$ENV{system_obj}->open_popup(
						url => $ENV{COMPREP}->link(
							alias	=> 'payment_secure_storage_skey_popup',
						),
						width => 600,
						heigth => 480,
					);
				}
				sw_log ("update_docs: process apply payment method to doc error=" . $error);
			} catch HSPC::Exc::Fatal with {
				$error = shift;
				sw_log ("update_docs: process apply payment method to doc error=" . $error);
			};
		} else {
			## remove from card
			my $ar_doc_rb = HSPC::MT::CCP->find_ar_doc_rb(id=>$rec->[0]);
			## not cc for ar_doc
			next unless $ar_doc_rb;
			## paid by another cc
			next unless ( $ar_doc_rb->method_id() == $paymethod->id() );
			$error = $ar_doc_rb->delete();
		}

		if ( $error ) {
			sw_log ("Can't perfom apply_paymethod_to_doc() for doc #".$rec->[0]);
			$sargs->page->add_error(string => $error);
			return $self->teaser_edit_docs();
		}
	}
	return $sargs->page->redirect('upd_ok_lc');
}

sub process_create_paymethod {
	my $self = shift;
	my %h = (
		template_id => undef,
		is_store    => undef,
		form_data   => undef,
		plugin_id   => undef,
		@_
	);

	my $trace = sw_atrace();

	my $errors;
	my $sargs = $self->session_args;

	if ( sw_argparam('item_cancel') ) {
		return $sargs->page->redirect();
	}

	unless ($sargs->paymethod_obj()) {
		my $engine = HSPC::MT::PM::Tools->get_mt_engine(
			template_id => $sargs->template_id,
		);
		eval "use $engine;";
		throw_exception(
			type        => 'error',
			log_message => "Cannot load required MT engine: $@"
		) if $@;

		$sargs->paymethod_obj($engine->new( template_id => $sargs->template_id ));
	}
	my $trans = atrans();

	my $exception_rised = 0;
	my ($error_type, $message, $log_message, $user_message, $vendor_message);

	try {
		sw_log ('Collecting payment method data...');
		$self->collect_data(
			paymethod        => $sargs->paymethod_obj(),
			is_admin         => $sargs->is_admin,
			account_type     => $sargs->account_type,
			effective_acc_no => $sargs->account_no,
			template_id      => $sargs->template_id,
			is_store         => $h{is_store},
			paymethod_acc_no => $sargs->paymethod_acc_no,
			form_data        => $h{form_data},
		);
		sw_log ('Collecting completed. Saving payment method...');
		## Even if secret key is absent private information can be saved.
		## No public_only parameter here.
		## Provide plugin_id parameter for check if that payment method type
		## is supported by this plugin. If undef then check will be performed
		## on all configured payment method types
		$sargs->paymethod_obj->save( plugin_id => $h{plugin_id});
		sw_log ('Payment method has been saved');
	} catch HSPC::Exception::Plugin::GeneralError with {
		my $e = shift;
		$exception_rised = 1;

		$error_type     = 'error';
		$log_message    = $e->log_message;
		$user_message   = $e->user_message;
		$vendor_message = $e->vendor_message;

		$sargs->page->add_error( string => ''.$e->log_message )
			unless ($sargs->internal());
	} catch HSPC::Exception::Plugin::InvalidParameter with {
		my $e = shift;
		$exception_rised = 1;

		$error_type = 'parameter';
		$message = $e->message;

		$sargs->page->add_error( string => ''.$e->message )
			unless ($sargs->internal());
	};

	if ( $exception_rised ) {
		$trans->rollback;

		throw_exception(
			type           => $error_type,
			message        => $message,
			log_message    => $log_message,
			user_message   => $user_message,
			vendor_message => $vendor_message
		) if ($sargs->internal());

		return $self->teaser_add();
	}

	$trans->commit;

	## no redirect if called for internal use
	return undef if ($sargs->internal());

	## Check if can pay documents
	my $plugins = HSPC::MT::PP->get_pp_plugin_info(
		is_active  => 1,
		is_direct  => 1,
		sub_type   => $sargs->paymethod_obj->plugin_type,
		vendor_id  => $sargs->paymethod_obj->vendor_id,
	);

	if ( @$plugins && $self->can_pay_docs(paymethod => $sargs->paymethod_obj) ) {
		## Paymethod added and can assing docs
		return $sargs->page->redirect(
		'add_ok_lc',undef,"/act,edit_docs/pmethod_id," . $sargs->paymethod_obj->id()) 
	}

    return $sargs->page->redirect('add_ok_lc'); ## redirect to listing by default
}

## >>  class
## =>> obj : reference to payment method object
## =>> is_admin
## => skip_acc_checking => 0|1 skip checking 
## account (usefull when account not created yet). 
sub collect_data {
	my $self = shift;
	my %h = (
		paymethod         => undef,
		template_id       => undef,
		is_admin          => undef,
		account_type      => undef,
		effective_acc_no  => undef,
		is_store          => undef,
		paymethod_acc_no  => undef,
		skip_acc_checking => undef,
		form_data         => undef,
		@_
	);

  	my $trace = sw_atrace();
	my $paymethod = $h{paymethod} ||
		throw_exception(
			type        => 'error',
			log_message => 'The paymethod parameter is mandatory'
		);

	my $template_id = $h{template_id} || $paymethod->template_id;
	throw_exception(
		type        => 'error',
		log_message => 'The template_id parameter is mandatory'
	) unless $template_id;

	my @err;
	my $is_store   = $h{is_store};
	my $account_no = sw_argparam('account_no') || $h{paymethod_acc_no} || $paymethod->account_no;

	throw_exception(
		type    => 'parameter',
		message => {
			string_id => 'acnt_not_exist',
			string_args => { id => "''" }
		}
	) unless ($account_no);

	my $account_info = HSPC::MT::Core->get_account_info( id => $account_no );

	my $vendor;

	if(!$h{skip_acc_checking}) {
		throw_exception(
			type    => 'parameter',
			message => {
				string_id => 'acnt_not_exist',
				string_args => { id => $account_no }
			}
		) unless ($account_info);
	
		$paymethod->account_no($account_no);

		## if current account NOT customer - check the rights to create card
		## if we are not customer - check is it possible to create CC ?
		my $check_account_type;
		my $check_account_no;

		if ($h{paymethod_acc_no}) {
			$check_account_type = $account_info->{type};
			$check_account_no   = $account_no;
		}
		else {
			$check_account_type = $h{account_type};
			$check_account_no   = $h{account_no};
		}

		if ( $check_account_type != &SW_CUSTOMER ) {
			## if we are reseller and this is card for himself - 
			## skip checking and allow. If not - check is we are vendor
			## for the target account ?
			if ($check_account_type != &SW_RESELLER || $account_no != $check_account_no) {
				if ( $account_info->{vendor_id} ne $check_account_no ) {
					throw_exception(
						type    => 'parameter',
						message => { string_id => 'not_vendor_no_card' }
					) unless ($account_info);
				}
			}
		}
 		$vendor=$account_info->{vendor_id};
  	}
 	else {
 		## account are able create only provider or reseller or any like 
		if ($h{paymethod_acc_no}) {
			$vendor = $h{paymethod_acc_no};
		}
		else {
			$vendor = $h{effective_acc_no};
		}
 	}


	## check security key is present for current provider(reseller)
	my $pr_config = HSPC::MT::PM::Tools->get_vendor_config(
		vendor_id => $vendor
	); 	

	throw_exception(
		type    => 'parameter',
		message => {string_id => 'not_vendor_no_card'}
	) unless($pr_config);

	if (!$pr_config->key1 && !$pr_config->key2
		&& $pr_config->card_proc_type == &SW_GM_CCARD_PRTYPE_MANUAL
	) {
		throw_exception(
			type    => 'parameter',
			message => {string_id => 'no_sk_in_system'}
		) unless($pr_config);
	}

	## Reseller can edit status of customer's payment method
	if ( $h{is_admin} || (
			$h{account_type} == &SW_RESELLER &&
			$h{effective_acc_no} == $vendor
		)
	) {
		$paymethod->status(&sw_argparam('status'));
	} else {
		$paymethod->status(&SW_PM_STATUS_UNCHECKED)
			unless defined $paymethod->status();
	}

	my $is_enabled = $is_store ? 1 : sw_argparam('is_enabled');
	$paymethod->enabled($is_enabled) if $h{is_admin} ||
		!$paymethod->id || $paymethod->status == &SW_PM_STATUS_OK ||
			$paymethod->status == &SW_PM_STATUS_UNCHECKED;

	$paymethod->vendor_id($vendor);

	## administrator managed properties
	if( $h{is_admin} || (
			$h{account_type} == &SW_RESELLER &&
			$h{effective_acc_no} == $vendor
		)
	) {
		$paymethod->recurring(&sw_argparam('rec_billing'));
		my $ip = &sw_argparam('ip_address');
		if ($ip) {
			if(HSPC::DataCheck::is_ip($ip)) {
				$paymethod->customer_ip($ip);
			} else {
				throw_exception(
					type    => 'parameter',
					message => {string_id => 'invalid_addr_specified'}
				);
			}
		}
	} else {
		$paymethod->recurring( defined &sw_argparam('rec_billing') ?
			&sw_argparam('rec_billing') : $pr_config->is_auto_recbill() );
		$paymethod->customer_ip( $h{form_data}->{ip_address} || $ENV{REMOTE_ADDR} )
			unless ( defined $paymethod->customer_ip );
	}

	my $is_new_card = $paymethod->id ? 0 : 1;

	my $decrypted = 0;

	## private data
	if (!$is_new_card) {
		$decrypted = $paymethod->decrypt();
	}

	my $old_expire_date = $paymethod->expire_date();

	my $template = HSPC::MT::PluginConfig->find_presentation_template(
		plugin_type => 'PM',
		template_id => $template_id,
	);

	HSPC::MT::PluginTools->init_plugin_argparams(
		plugin_argparams => $h{form_data},
	) if $h{form_data};

	eval "use $template;";
	throw_exception(
		type        => 'error',
		log_message => "Cannot load template for $template_id: $@"
	) if $@;

	my $keep_secure_code = $self->__keep_secure_code(vendor_id => $vendor);

	my $data = $template->collect_data(
		public_data      => $paymethod->public_data,
		account_data     => $account_info,
		secure_data      => $decrypted ? $paymethod->secure_data : undef,
		only_public      => (!$h{is_admin} && !$is_new_card),
		keep_secure_code => $keep_secure_code,
	);

	my $mt_template = $paymethod->template;

	eval "use $mt_template;";

	throw_exception(
		type        => 'error',
		log_message => "Cannot load MT template for $template_id: $@"
	) if $@;

	$paymethod->public_data($mt_template->get_public_data( data => $data))
		if ($mt_template->can('get_public_data'));

	if ( $mt_template->can('get_expiration_date') ) {
		$paymethod->expire_date($mt_template->get_expiration_date(
			public_data => $paymethod->public_data,
		));
	} else {
		$paymethod->expire_date('');
	}

	## If customer added payment method with incorrect expiration date and it
	## become expired then set "Unchecked" status when customer updates expiration
	## date to corect			
	if (
		!$h{is_admin}
		&& $paymethod->status() == &SW_PM_STATUS_EXPIRED
		&& $paymethod->expire_date ne $old_expire_date
	) {
		$paymethod->status(&SW_PM_STATUS_UNCHECKED)
	}

	$paymethod->secure_data($mt_template->get_secure_data(
		data             => $data,
		keep_secure_code => $keep_secure_code,
	)) if ($mt_template->can('get_secure_data'));

	throw_exception(
		type        => 'error',
		log_string => "The $template_id template does not have get_paymethod_type method defined"
	) unless $mt_template->can('get_paymethod_type');

	$paymethod->type($mt_template->get_paymethod_type(
		secure_data => $paymethod->secure_data,
		public_data => $paymethod->public_data,
	));

	unless (defined $paymethod->name) {

		throw_exception(
			type       => 'error',
			log_string => "The $template_id template does not have get_public_number method defined"
		) unless $mt_template->can('get_public_number');

		my $public_number = $mt_template->get_public_number(
			secure_data => $paymethod->secure_data,
			public_data => $paymethod->public_data,
		);

		$paymethod->name("*****$public_number");
	}

	$trace->add("ok");
	return undef;
}

## >> class 
## =>> cc - optional CC obj to update
## + same arguments as for "teaser"
sub process_update_paymethod {
    my $self = shift;

	my $trace = sw_atrace();
	my $sargs = $self->session_args;

	if ( sw_argparam('item_cancel') ) {
		return $self->teaser_view();
	}

	unless ( $sargs->paymethod_obj && $sargs->paymethod_obj->id ) {
		sw_die(__PACKAGE__ . ": Paymethod object is undefined");
	}

	if (   $sargs->paymethod_obj->vendor_id != $sargs->account_no
		&& $sargs->paymethod_obj->account_no != $sargs->account_no )
	{
		$ENV{system_obj}->exit_noobject();
	}

	my $exception_rised = 0;

	if(sw_argparam('item_submit')) {
		my $trans = atrans();
		try {
			sw_log ("Collecting payment method data...");

			$self->collect_data(
				paymethod        => $sargs->paymethod_obj(),
				is_admin         => $sargs->is_admin,
				account_type     => $sargs->account_type,
				effective_acc_no => $sargs->account_no,
				template_id      => $sargs->template_id,
				is_store         => undef,
				paymethod_acc_no => $sargs->paymethod_acc_no,
			);

			sw_log ('Collecting completed. Saving payment method...');
			## Set only_public parameter here only if this method has been called by customer.
			## Private data update is available to a vendor only
			$sargs->paymethod_obj->save();
			sw_log ('Payment method has been saved' );
		} catch HSPC::Exception::Plugin::GeneralError with {
			my $exception = shift;
			$exception_rised = 1;

			$trans->rollback;

			throw HSPC::Exception::Plugin::GeneralError(
				string_id => $exception->{string_id}, string_params => $exception->{string_params}
			) if ($sargs->internal());

			$sargs->page->add_error(
				string => string($exception->{string_id}, $exception->{string_params})
			);
		} catch HSPC::Exception::Plugin::InvalidParameter with {
			my $exception = shift;
			$exception_rised = 1;

			$trans->rollback;

			throw HSPC::Exception::Plugin::InvalidParameter(
		    	message   => $exception->message,
			) if ($sargs->internal());

			$sargs->page->add_error( string => $exception->message );

		} catch HSPC::Exception::Plugin::SecureKey with {
			$trans->rollback;
			$exception_rised = 1;

			$ENV{system_obj}->open_popup(
				url => $ENV{COMPREP}->link(
					alias	=> 'payment_secure_storage_skey_popup',
				),
				width => 600,
				heigth => 480,
			);
		};

		if ( $exception_rised ) {
			return $self->teaser_edit();
		}

		$trans->commit;

		return $sargs->page->redirect('upd_ok_lc');
	}
	return undef;
}

sub process_delete_paymethod {
    my $self = shift;

	my $trace = sw_atrace();
	my $sargs = $self->session_args;

	unless ( $sargs->paymethod_obj && $sargs->paymethod_obj->id ) {
		sw_die(__PACKAGE__ . ": Paymethod object is undefined");
	}

	my $errors;
	my $trans = atrans();
	try {
		$errors = $sargs->paymethod_obj->delete();
	} catch HSPC::Exception::Plugin::SecureKey with {
		$errors = shift;
		$ENV{system_obj}->open_popup(
			url => $ENV{COMPREP}->link(
				alias	=> 'payment_secure_storage_skey_popup',
			),
			width => 600,
			heigth => 480,
		);
	} catch HSPC::Exc::Fatal with {
		$errors = shift;
	};

	if ( $errors ) {
		sw_log (__PACKAGE__ . 
			"process_update_paymethod: can not delete payment method. Error: "
			. $errors
		);
		$trans->rollback;
		## Object has to be reloaded as some data is incorrect now
		$sargs->paymethod_obj->load( id => $sargs->paymethod_obj->id)
			if ($sargs->paymethod_obj->id);
		$sargs->page->add_error( string => $errors );
		return $self->teaser_view();
	}

	$trans->commit;

	## redirect to listing if OK
	return $sargs->page->redirect('delete_ok_lc', undef, undef);
}

###############################################################################
## Visual PayMethod Plugins API
###############################################################################
## >>  class
## =>> page
## =>> obj : reference to payment method object
## => is_admin
## => internal
## => is_order_view : show a bit customised form, on that huge Order Processing Screen
## => skey_id, skey, pass : security data
## => cc_num_link : show paymethod number as link to url defined as value of this parameter
## < undef
sub view_form    {
	my $class = shift;
	my %h = (
		obj        => undef,
		page       => undef,
		is_admin   => undef,
		gap_height => undef,
		with_link  => undef,
		@_
	);

	my $page = $h{page} || sw_die('page=> undefined');
	sw_die('obj=> undefined') unless $h{obj};
	
	my $trace = sw_atrace();
	
	## Decrypt card if we can
	my $decrypted = 0;

	try {
		$decrypted = $h{obj}->decrypt();
	} catch HSPC::Exception::Plugin::SecureKey with {
		## Do nothing here because as we will show public info
	} otherwise {
		my $e = shift;
		## Do nothing here because as we will show public info
		if (ref($e) =~ /^HSPC::Exception::Plugin/) {
			sw_log_warn('can not decrypt payment method with error: ' . $e->log_message);
			$page->add_error(string => $e->vendor_message );
		} else {
			$page->add_error(string => $e );
		}
		$page->statuses();
	};

	my ($type_info) = @{HSPC::MT::PM::Tools->get_pm_template_description(types => [$h{obj}->type])};
	my $template = HSPC::MT::PluginConfig->find_presentation_template(
		plugin_type => 'PM',
		template_id => $h{obj}->template_id,
	);

	my $result;

	eval "use $template;";

	unless ($@) {
		set_request_data('plugin_template_id', $h{obj}->template_id);
		set_request_data('plugin_type', 'PM');
		set_request_data('plugin_vendor_id', $h{obj}->vendor_id);
		if ($template->can('view_form')) {
			$result = $template->view_form(
				expire_date      => $h{obj}->expire_date,
				keep_secure_code => $class->__keep_secure_code(vendor_id => $h{obj}->vendor_id),
				type             => $h{obj}->type,
				name             => $h{obj}->name,
				is_admin         => $h{is_admin},
				type_info        => $type_info,
				public_data      => $h{obj}->public_data,
				secure_data      => $decrypted ? $h{obj}->secure_data : undef,
				number_link      => $h{with_link} ? $ENV{COMPREP}->link(
					alias => 'payment_methods_'.$h{obj}->plugin_type, {
						act         => 'view',
						pmethod_id  => $h{obj}->id,
					}
				) : undef,
			);
		} else {
			sw_log_warn("PLugin $template does not have the view_form method defined");
		}
	} else {
		sw_log_warn("Can not load plugin $template: $@");
	}

	$page->edit_table_break(
		new_table_class => '',
		cellpadding     => 0,
		height          => defined $h{gap_height} ? $h{gap_height} : 13,
		cell_open       => 1,
	);
	$page->{edit_out} .= $result->{content};
	$page->edit_table_break(
		height     => 1,
		cell_close => 1,
	);

	$page->view_info_folder( title_id => 'paymethod_general_data_uc' );

	## Card status
	$page->view_info_text(
		title_id => 'cc_status', as_plain=>0,
		value => $class->colored_status($h{obj}->status)
	);

	## Account Name
	if (!$h{is_order_view} && $ENV{security_obj}->account_type != &SW_CUSTOMER) {
		my $acct_obj = HSPC::MT::Core->find_customer(id=>$h{obj}->account_no);
		if ($acct_obj) {
			my $link;
			if ( $acct_obj->id() != $ENV{security_obj}->account_no() ) {
				$link =	$ENV{COMPREP}->link(
					alias => $acct_obj->type == SW_RESELLER
					? 'resellers_view'
					: 'customers_view',
					{ id => $h{obj}->account_no }
				);
			} else {
				$link = $ENV{COMPREP}->link(alias => 'cc_core_front_tab_sum');
			}
 
			$page->view_info_text(title_id=>'accnt_uc',
				value=>'#' . $h{obj}->account_no . '&nbsp;' . $acct_obj->name,  type=>'char',
				link_url => $link
			);
		}
	}

	## Availability status
	$page->view_info_text(
		title_id => 'allows_payment_paymethod',
		value    => $h{obj}->enabled,
		type     => 'bool'
	);

	## Use for recurrency billing status
	$page->view_info_text(
		title_id => 'use_4rec_paying',
		value    => $h{obj}->recurring,
		type     => 'bool'
	);
	## Credit Card IP Address
 	if ( $h{obj}->customer_ip && $h{is_admin} || (
			$h{account_type} == &SW_RESELLER &&
			$h{effective_acc_no} != $h{obj}->account_no
		)
	) {
		$page->view_info_text(
			title_id => 'customer_ip', as_plain=>0,
			value => $class->display_ip($h{obj}->customer_ip)
		);
	}

	return undef;
}


sub add_form     {
	my $class = shift;
	my %h    = (
		page                  => undef,
		obj                   => undef,
		account_no            => undef,
		show_account_selector => undef,
		is_admin              => undef,
		template_id           => undef,
		vendor_id             => undef,
		plugin_id             => undef,
		@_
	);

	my $trace = sw_atrace();

	my $show_account_selector = $h{show_account_selector};
	my $is_admin              = $h{is_admin} ? 1 : 0;
	my $account_no            = $h{obj} ? $h{obj}->account_no : $h{account_no};
	my $page                  = $h{page};
	my $vendor_id             = $h{vendor_id};
	my $plugin_id             = $h{plugin_id};
	my $is_store              = $plugin_id ? 1 : 0;

	throw HSPC::Exception::Plugin::GeneralError(
		log_message => 'no page'
	) unless ( $page || $is_store );

	throw HSPC::Exception::Plugin::GeneralError(
		log_message => ' no template_id'
	) unless ( $h{template_id} );

	unless ($h{obj}) {
		my $engine = HSPC::MT::PM::Tools->get_mt_engine(
			template_id => $h{template_id},
		);
		eval "use $engine;";
		throw HSPC::Exception::Plugin::GeneralError(
			log_message => "Can not load MT engine: $@"
		) if ( $@ );

		$h{obj} = $engine->new( template_id => $h{template_id} );
	}

	my $account_info;
	if ($account_no) {
		$account_info = HSPC::MT::Core->get_account_info( id => $account_no );
	}

	if ( $show_account_selector && $class->popup_url() ) {
		$page->edit_view_hidden(view_name => 'just_reload', value => 0);
		$page->edit_view_popup(
			title_id      => 'accnt_uc',
			view_name     => 'account_no',
			value         => $account_no,
			value_title   => $account_info ? $account_info->{name} : '',
			is_show_value => 1,
			popup_url     => $class->popup_url(),
			not_empty     => 1
		);
	}

	my $template = HSPC::MT::PluginConfig->find_presentation_template(
		plugin_type => 'PM',
		template_id => $h{template_id},
	);

	eval "use $template;";
	throw HSPC::Exception::Plugin::GeneralError(
		log_message => "Can not load presentation engine: $@"
	) if ( $@ );

	my $result;
	if ($template->can('add_form')) {
		## not migrated
		my $allowed_types = HSPC::MT::PM::Tools->pm_allowed_types(
			template_id => $h{template_id},
			vendor_id   => $vendor_id,
			plugin_id   => $plugin_id,
		);
		unshift @$allowed_types, ['none', string('autodetect')];

		set_request_data('plugin_type', 'PM');
		set_request_data('plugin_template_id', $h{template_id});
		set_request_data('plugin_vendor_id', $vendor_id);
		$result = $template->add_form(
				keep_secure_code => $class->__keep_secure_code(vendor_id => $vendor_id),
				public_data      => $h{obj}->public_data,
				secure_data      => $h{obj}->secure_data,
				allowed_types    => $allowed_types,
				account_data     => $account_info,
				type             => $h{obj}->type,
				expire_date      => $h{obj}->expire_date,
		);
		$result->{content} .= qq{
			<input type="hidden" id="template_id" name="template_id" value="$h{template_id}">
		};
	} else {
		throw HSPC::Exception::Plugin::GeneralError(
			log_message => "Template $h{template_id} does not have the add_form method defined"
		);
	}

	if ($is_store) {
		return $result;
	} else {
		$page->edit_table_break(
			new_table_class => '',
			cellpadding     => 0,
			height          => 13,
			cell_open       => 1,
		);
		$page->row_open();
		$page->{edit_out} .= $result->{content};
		$page->edit_table_break(
			height     => 1,
			cell_close => 1,
		);
		$page->row_close();

		$page->edit_view_info( value => string('paymethod_general_data_uc') );

		if($account_info) {
			$page->edit_view_text(title_id=>'accnt_uc', value=>$account_info->{name});
		}

  		if ( $is_admin || ( $account_no &&
				$h{account_type} == &SW_RESELLER &&
				$h{effective_acc_no} != $account_no
			)
		) {
			my $cc_status = &SW_PM_STATUS_TXT;
			$cc_status = [
				map { [$_, $cc_status->{$_}] }
					grep { $_ != &SW_PM_STATUS_DELETED } keys %$cc_status
			];
			$page->edit_view_combo(
				title_id  => 'status',
				view_name => 'status',
				value     => ($h{obj}->status || &SW_PM_STATUS_UNCHECKED),
				options   => $cc_status,
				not_empty => 1
			);
		}

		$page->edit_view_check(
			title_id  => 'allows_payment_paymethod',
			view_name => 'is_enabled',
			value     => defined $h{obj}->enabled() ? defined $h{obj}->enabled() : 1,
		);
		$page->edit_view_check(
			title_id  => 'use_4rec_paying',
			view_name => 'rec_billing',
			value     => defined $h{obj}->recurring() ? $h{obj}->recurring() : $account_info,
		);

		## Customer IP (needed by WFSS)
  		if ( $is_admin || ( $account_no &&
				$h{account_type} == &SW_RESELLER &&
				$h{effective_acc_no} != $account_no
			)
		) {
			$page->edit_view_input(
				title_id   => 'customer_ip',
				view_name  => 'ip_address',
				value      => $h{obj}->customer_ip || 
					$account_info->{ip_address},
				not_empty  => 1,
				max_length => 15
			);
		}
	}

	return $result->{js_onsubmit};
}

sub edit_form    {
	my $class = shift;
	my %h = (
		obj              => undef,
		page             => undef,
		is_admin         => undef,
		account_type     => undef,
		effective_acc_no => undef,
		@_
	);

	my $page = $h{page} || &sw_die('page=> undefined');
	my $paymethod = $h{obj}  || $h{card} || &sw_die('obj=> undefined');

	my $trace = sw_atrace("account_no=>$h{account_no}");
	my $acc;

	## Decrypt card if we can
	my $decrypted = 0;

	try {
		$decrypted = $paymethod->decrypt();
	} catch HSPC::Exception::Plugin::SecureKey with {
		## Do nothing here because as we will show public info
	} otherwise {
		my $exc = shift;
		## Do nothing here because as we will show public info
		sw_log_warn("can not decrypt payment method with error = $exc");
		$page->add_error(string => string($exc->{string_id}, $exc->{string_param} ? %{$exc->{string_param}} : undef) );
	};

	my $template = HSPC::MT::PluginConfig->find_presentation_template(
		plugin_type => 'PM',
		template_id => $paymethod->template_id,
	);

	my $result;

	eval "use $template;";
	unless ($@) {
		if ($template->can('edit_form')) {
			my $allowed_types = HSPC::MT::PM::Tools->pm_allowed_types(
				template_id => $paymethod->template_id,
				vendor_id   => $paymethod->vendor_id,
			);
			unshift @$allowed_types, ['none', string('autodetect')];

			set_request_data('plugin_template_id', $paymethod->template_id);
			set_request_data('plugin_type', 'PM');
			set_request_data('plugin_vendor_id', $paymethod->vendor_id);
	
			$page->edit_table_break(
				new_table_class => '',
				cellpadding     => 0,
				height          => 13,
				cell_open       => 1,
			);
			$page->row_open();
			$result = $template->edit_form(
					keep_secure_code => $class->__keep_secure_code(vendor_id => $h{obj}->vendor_id),
					expire_date      => $h{obj}->expire_date,
					name             => $h{obj}->name,
					type             => $h{obj}->type,
					is_admin         => $h{is_admin},
					allowed_types    => $allowed_types,
					public_data      => $h{obj}->public_data,
					secure_data      => $decrypted ? $h{obj}->secure_data : undef,
			);

			$page->{edit_out} .= $result->{content};
			$page->edit_table_break(
				height     => 1,
				cell_close => 1,
			);
			$page->row_close();
		} else {
			sw_log_warn("Plugin $template does not have the edit_form method defined");
		}
	} else {
		sw_log_warn("Can not load plugin $template: $@");
	}

	$h{page}->edit_view_info( value => string('paymethod_general_data_uc') );

	unless($h{account_no}) {
		$acc =
		  HSPC::MT::Core->find_account( account_id => $paymethod->account_no );
		$page->edit_view_text(title_id=>'accnt_uc', value=>$acc->name);
	}

	## Reseller can edit status of customer's payment method
	if ( $h{is_admin} || (
			$h{account_type} == &SW_RESELLER &&
			$h{effective_acc_no} != $paymethod->account_no
		)
	) {
		my $cc_status = &SW_PM_STATUS_TXT;
		$cc_status = [ map { [ $_, $cc_status->{$_} ] }
						   grep {$_ != &SW_PM_STATUS_DELETED} keys %$cc_status ];
		$page->edit_view_combo(
			title_id  => 'status',
			view_name => 'status',
			value     => ( $paymethod->status || &SW_PM_STATUS_UNCHECKED ),
			options   => $cc_status,
			not_empty => 1
		);
	};

	$page->edit_view_check(
		title_id  => 'allows_payment_paymethod',
		view_name => 'is_enabled',
		value     => $paymethod->enabled
	);

	$page->edit_view_check(
		title_id  => 'use_4rec_paying',
		view_name => 'rec_billing',
		value     => $paymethod->recurring
	);

	## ip_address
 	if ( $h{is_admin} || (
			$h{account_type} == &SW_RESELLER &&
			$h{effective_acc_no} != $paymethod->account_no
		)
	) {
		if (!$acc) {
			$acc =
			  HSPC::MT::Core->find_account(
				account_id => $paymethod->account_no );
		}
		$page->edit_view_input(
			title_id   => 'customer_ip',
			view_name  => 'ip_address',
			value      => $paymethod->customer_ip || $acc->ip_address(),
			not_empty  => 1,
			max_length => 15
		);
	}

	return $result->{js_onsubmit};
}

sub build_path {
	my $self = shift;
	my %h = (
		step => undef,
		@_
	);

	my $trace = &sw_atrace("step => $h{step}");
	$h{step} = 1 unless ($h{step});

	if ( $self->session_args->bread_crumbs->{$h{step}} ) {

		$self->session_args->page->add_header_nav_path(
			path => {
				title_id => $self->session_args->bread_crumbs->{$h{step}}->{title_id},
				url => eval($self->session_args->bread_crumbs->{$h{step}}->{url})
			},
			force => 1
		);
	}

	$h{step}++;
	$trace->add("Res = $h{step}");
	return $h{step};
}

sub get_screen_id {
	my $self = shift;
	my %h = (
		page => undef,
		@_
	);

	sw_die(__PACKAGE__." : invalid object method call")
		unless ref($self);

	sw_die(__PACKAGE__." : The page parameter must be provided")
		unless( $h{page} );
	
	my $screen_id = $self->session_args->screen_ids->{$h{page}};

	return $screen_id ? $screen_id : 'xx.xx';
}

sub teaser_edit {
	my $self = shift;

	my $trace = sw_atrace();

	my $sargs         = $self->session_args;
	my $paymethod = $sargs->paymethod_obj();

	unless ( $paymethod && $paymethod->id ) {
		sw_die(__PACKAGE__ . ": Paymethod object is undefined");
	}

	if ( $paymethod->vendor_id != $sargs->account_no &&
		$paymethod->account_no != $sargs->account_no
	) {
		$ENV{system_obj}->exit_noobject();
	}

	if (sw_argparam('item_cancel')) {
		return $sargs->page->redirect();
	} elsif (sw_argparam('item_delete') && $paymethod->id) {
		return $self->process_delete_paymethod();
	}

	my $main_page = $sargs->page;

	if ( $sargs->draw_header() ) {
		## Backward compatibility for CP that's not migrated to COMPREP yet
		if ($ENV{COMPREP}) {
			$ENV{COMPREP}->node(edit => 1)->{title_hash}->{title} =
				string('paymethod') . ' #' . $paymethod->id;
			$main_page->title(pages => []);
		} else {
			$self->build_path( step => 3 );

			$main_page->header();
			$main_page->title(
				'title' => string('paymethod')." #".$paymethod->id,
				'devscreen_id' => $self->get_screen_id( page => 'edit' ),
			);
			$main_page->tab();
		}
	}
	$main_page->statuses();

	my $url_ext = "/act,update/pmethod_id,".$paymethod->id;
	$main_page->edit_open(url_ext=>$url_ext,disable_header=>1);

	## draw account general settings edit screen
	my $js_code = $self->edit_form(
		obj              => $paymethod,
		page             => $main_page,
		is_admin         => $sargs->is_admin,
		account_type     => $sargs->account_type,
		effective_acc_no => $sargs->account_no,
	);

	$main_page->edit_close(
		upd_id      => 'ok',
		code_custom => $js_code 
	);
	$main_page->footer();

	return undef;
}

sub get_help_url {
	my $self = shift;

	my $sargs = $self->session_args;

	my %action_map = (
		view      => 'view',
		edit      => 'edit',
		add       => 'add',
	);

	my $action = $action_map{$sargs->action};
	return undef unless $action;

	my $template_id = $sargs->template_id ||
		($sargs->paymethod_obj && $sargs->paymethod_obj->template_id);

	my $url;
	if ( $template_id ) {
		my $template = HSPC::MT::PluginConfig->find_presentation_template(
			plugin_type => 'PM',
			template_id => $template_id,
		);

		eval "use $template;";

		unless ($@) {
			my $help;
			if ($template->can('get_help_page')) {
				$help = $template->get_help_page(
					action        => $action,
					customer_view => $sargs->is_admin ? 0 : 1,
					language      => uc(current_lang()),
				);
			}

			if ($help) {
				set_request_data('plugin_type', 'PM');
				$url = HSPC::PluginToolkit::General::get_help_url(
					action        => $action,
					customer_view => $sargs->is_admin ? 0 : 1,
					template_id   => $template_id,
				);
			}
		}
	}

	return $url;
}

sub get_help_page {
	my $class = shift;
	my %h = (
		action        => undef,
		template_id   => undef,
		language      => undef,
		customer_view => undef,
		@_
	);

	my $template = HSPC::MT::PluginConfig->find_presentation_template(
		plugin_type => 'PM',
		template_id => $h{template_id},
	);

	eval "use $template;";

	my $help;
	unless ($@) {
		if ($template->can('get_help_page')) {
			$help = $template->get_help_page(
				action        => $h{action},
				customer_view => $h{customer_view},
				language      => $h{language},
			);
		}
	}

	return $help;
}

sub get_tooltip {
	my $self = shift;

	my $sargs = $self->session_args;

	my %action_map = (
		view      => 'view',
		edit      => 'edit',
		add       => 'add',
	);

	my $action = $action_map{$sargs->action};
	return undef unless $action;

	my $template_id = $sargs->template_id ||
		($sargs->paymethod_obj && $sargs->paymethod_obj->template_id);

	my $tooltip;
	if ( $template_id ) {
		my $template = HSPC::MT::PluginConfig->find_presentation_template(
			plugin_type => 'PM',
			template_id => $template_id,
		);

		eval "use $template;";

		unless ($@) {
			if ($template->can('get_tooltip')) {
				$tooltip = $template->get_tooltip(
					action        => $action,
					customer_view => !$sargs->is_admin,
					language      => uc(current_lang()),
				);
			}
		}
	}

	return $tooltip;
}

sub __keep_secure_code {
	my $class = shift;
	my %h = (
		vendor_id => undef,
		@_
	);

	if ($h{vendor_id}) {
		return HSPC::MT::Billing::Setup->is_save_cvc(
			provider_id => $h{vendor_id}
		);
	}

	return undef;
}

sub __title_single_id {
	return 'paymethod';
}

sub __title_plural_id {
	return 'paymethods_uc';
}

1;
